import time
import math
import os
import datetime

from generic_module import Module
from collections import deque
from lxml import etree
import threading
import logging
from py_logging.cestel_logging import log, log_exception

class AggregationModule(Module):
  def __init__(self, name, matching_timeout, fuzziness, save_to_disk, boss_receive, root_path):
    Module.__init__(self, name, root_path)
    self.matching_timeout = matching_timeout
    self.fuzziness = fuzziness
    self.save_to_disk = save_to_disk
    self.timeout = 0
    self.boss_receive = boss_receive
    
    self.vehicles_dict = dict()
    self.deque_lock = threading.Lock()
    self.offsets_from_boss_dict = dict()
    
    self.swd_version = ""
    self.modules_per_lane = dict()
  
  def set_swd_version(self, version):
    self.swd_version = version
  
  def tell_offset(self, recv_name, offsets_dict):
    self.offsets_from_boss_dict[recv_name] = offsets_dict
    log(self.logger, 'offsets_from_boss_dict: {0}'.format(str(self.offsets_from_boss_dict)))
  
  def add_vehicle(self, vehicle, recv_module_name):
    sub_vehicle = vehicle.getchildren()[0]
    #in case of no lane data, lane = 0 (no actual lane will ever have the index 0)
    lane = "0"
    try:
      lane = sub_vehicle.find("lane").text
    except:
      log_exception(self.logger, 'Failed to obtain lane:')
    if lane not in self.modules_per_lane:
      self.modules_per_lane[lane] = set()
    self.modules_per_lane[lane].add(recv_module_name)
    #print etree.tostring(sub_vehicle)
    if recv_module_name not in self.vehicles_dict:
      self.vehicles_dict[recv_module_name] = dict()
    if lane not in self.vehicles_dict[recv_module_name]:
      self.vehicles_dict[recv_module_name][lane] = deque()
    with self.deque_lock:
      self.vehicles_dict[recv_module_name][lane].append(sub_vehicle)
        
  #--- "private"
  def find_and_form(self, boss_receive_sub_veh):
    ts = boss_receive_sub_veh.find("ts").text
    start = time.mktime(time.strptime(ts, "%Y-%m-%d-%H-%M-%S-%f"))
    #---
    vehicle = etree.Element("vehicle")
    vehicle.append(boss_receive_sub_veh)
    #clear "dead" recvs
    
    #boss_lane
    lane = "0"
    try:
      lane = boss_receive_sub_veh.find("lane").text
    except:
      pass
    #print self.vehicles_dict
    for key in self.vehicles_dict:
      if key == self.boss_receive:
        continue
      try:
        with self.deque_lock:
          while "0" in self.vehicles_dict[key] and len(self.vehicles_dict[key]["0"]) > 0 and self.compare_timestamps(self.vehicles_dict[key]["0"][0], boss_receive_sub_veh, key) == 3:
            self.vehicles_dict[key]["0"].popleft()
          while lane != "0" and lane in self.vehicles_dict[key] and len(self.vehicles_dict[key][lane]) > 0 and self.compare_timestamps(self.vehicles_dict[key][lane][0], boss_receive_sub_veh, key) == 3:
            self.vehicles_dict[key][lane].popleft()
      except:
        log_exception(self.logger, 'Find and form error 1:')
        try:
          self.deque_lock.release()
        except:
          pass
        return

    found_list = list()
    found_list.append(self.boss_receive)
    #---
    #if both boss receive and guest receive event are on the same lane, use that
    #if either boss, or guest (don't), don't have lane defined, use lane "0"
    while True:
      #if we have found everything for this lane, break
      if len(found_list) == len(self.modules_per_lane[lane]):
        break
      for key, value in self.vehicles_dict.items():
        if key not in found_list:
          try:
            with self.deque_lock:
              if lane in value:
                log(self.logger, str(key) + " " + str(lane) + " " + str(len(value[lane])), sev=logging.DEBUG)
                for i in range(len(value[lane])):
                  self.logger.debug(str(value[lane]))
                  if self.compare_timestamps(value[lane][i], boss_receive_sub_veh, key) == 1:
                    vehicle.append(value[lane][i])
                    found_list.append(key)
                    del value[lane][i]
                    break
              elif lane != "0" and "0" in value and lane not in value:
                for i in range(len(value["0"])):
                  if self.compare_timestamps(value["0"][i], boss_receive_sub_veh, key) == 1:
                    vehicle.append(value["0"][i])
                    found_list.append(key)
                    del value["0"][i]
                    #broken = True
                    break
          except:
            log_exception(self.logger, 'Find and form error 2:')
            try:
              self.deque_lock.release()
            except:
              pass
            break
      if time.time() - start > self.matching_timeout:
        break
      self.zzzzz(0.5)
    return vehicle   
  
  def save_to_file(self, str_vehicle):
    #if site name is unknown at this point, we have nowhere to save
    if self.sname == "":
      return
    date = datetime.datetime.now().strftime('%Y-%m-%d')
    if not os.path.exists(self.sites_path + "/" + self.sname + "/ext"):
      os.mkdir(self.sites_path + "/" + self.sname + "/ext")
    fd = None
    if os.path.exists(self.sites_path + "/" + self.sname + "/ext/" + date + ".xml"):
      fd = open(self.sites_path + "/" + self.sname + "/ext/" + date + ".xml", "a+")
    else:
      first_line = '<?xml version="1.0" ?>\n<swd version=' + '"' + self.swd_version + '"><site><name>' + self.sname + '</name><vehicles>'
      fd = open(self.sites_path + "/" + self.sname + "/ext/" + date + ".xml", "a+")
      fd.write(first_line)
    if fd == None:
      return
    fd.seek(0, os.SEEK_END)
    pos = fd.tell() - 1
    #find the last vehicle closing tag
    while pos > 0 and fd.read(10) != "</vehicle>":
      pos -= 1
      fd.seek(pos, os.SEEK_SET)  
    #delete everything from the end of last vehicle closing tag onwards
    if pos > 0:
      fd.seek(pos+10, os.SEEK_SET)
      fd.truncate()
    fd.write("\n" + str_vehicle)
    last_line = "</vehicles></site></swd>"
    fd.write("\n" + last_line)
    fd.close()
  
  #module_name1 is the name of the "guest" module    
  def compare_timestamps(self, sub_veh1, sub_veh2, module_name1):
    try:
      ts1 = sub_veh1.find("ts").text
      ts2 = sub_veh2.find("ts").text
      v = 0
      try:
        v = float(sub_veh2.find("v").text)
      except:
        try:
          v = float(sub_veh1.find("v").text)
        except:
          pass
      l1 = sub_veh1.find("lane")
      add_to_offset = sub_veh1.find("siwim_i_add_to_offset")
      parts1 = ts1.split("-")
      parts2 = ts2.split("-")
      t_offset = 0
      factor = 1
      try:
        t_offset = self.offsets_from_boss_dict[module_name1][l1.text] / v
        try:
          factor = self.offsets_from_boss_dict[module_name1][l1.text] / abs(self.offsets_from_boss_dict[module_name1][l1.text])
          offoffset = float(add_to_offset.text)
          t_offset = t_offset + (offoffset * factor)
          if offoffset == 0:
            offoffset = (float(sub_veh2.find("whlbse").text) + float(sub_veh2.find("append").text)) / v
            t_offset = t_offset + offoffset
        except:
          log(self.logger, "offset:", sev=logging.DEBUG, exc_info=True)
      except:
        log_exception(self.logger, 'offsets_from_boss')
      timestamp1 = time.mktime(datetime.datetime.strptime(ts1, "%Y-%m-%d-%H-%M-%S-%f").timetuple()) + float(parts1[len(parts1) - 1]) / 1000 
      timestamp2 = time.mktime(datetime.datetime.strptime(ts2, "%Y-%m-%d-%H-%M-%S-%f").timetuple()) + float(parts2[len(parts2) - 1]) / 1000 + t_offset
      if math.fabs(timestamp1 - timestamp2) <= self.fuzziness:
        log(self.logger, 'matched ' + ts1 + ' with ' + ts2, sev=logging.DEBUG)
        log(self.logger, 'fuzz: ' + str(self.fuzziness), sev=logging.DEBUG)
        log(self.logger, str(t_offset), sev=logging.DEBUG)
        return 1
      else:
        if timestamp1 > timestamp2:
          return 2
        else:
          #my claim is that we should only clear timestamps (return 3) that are older than self.matching_timeout
          #twice that should suffice ... possibly ... maybe ...
          if math.fabs(timestamp1 - timestamp2) > (2 * self.matching_timeout):
            return 3
          else:
            return 4
    except:
      log_exception(self.logger, 'Compare timestamps error:')
      return 3
            
  def run(self):
    import time
    self.alive = True
    self.end = False
    log(self.logger, "Starting " + self.name + " aggregation module ...", sev=logging.DEBUG)
    try:
      while True:
        if self.end:
          self.alive = False
          log(self.logger, 'Thread closed correctly.', sev=logging.DEBUG)
          return   
        #8 lanes fapp
        for i in range(9):
          lane = str(i)
          if self.boss_receive in self.vehicles_dict and lane in self.vehicles_dict[self.boss_receive] and len(self.vehicles_dict[self.boss_receive][lane]) > 0:
            boss_receive_sub_veh = self.vehicles_dict[self.boss_receive][lane].popleft()
            vehicle = self.find_and_form(boss_receive_sub_veh)
            if self.save_to_disk == 1:
              vehicle_str = etree.tostring(vehicle).decode()
              self.save_to_file(vehicle_str)
            #print etree.tostring(vehicle)
            
            if vehicle.find("alpr") != None or vehicle.find("adr") != None:
              anpr_node = etree.Element("anpr")
              if vehicle.find("alpr") != None:
                anpr_node.append(vehicle.find("alpr").find("ts"))
                anpr_node.append(vehicle.find("alpr").find("lp"))
                vehicle.find("alpr").getparent().remove(vehicle.find("alpr"))
              if vehicle.find("adr") != None:
                anpr_node.append(vehicle.find("adr").find("hgp"))
                vehicle.find("adr").getparent().remove(vehicle.find("adr"))
              vehicle.append(anpr_node)
              
            for  key, mod in self.downstream_modules_dict.items():
              mod.add_vehicle(vehicle, self.name)
              mod.set_site_name(self.sname)
          #print etree.tostring(vehicle)
        time.sleep(0.5)
    except:
      log_exception(self.logger, 'Fatal exception in Aggregation Module:')
      self.alive = False
