import serial
import time
import socket
#import datetime
from py_logging.cestel_logging import log, log_exception
import logging
import os
import datetime

from generic_module import Module

class HistModule(Module):
  def __init__(self, name, udp_ip, udp_port, serial_port, baud, interval, root_path, relevant_for_listeners):
    Module.__init__(self, name, root_path)
    self.udp_ip = udp_ip
    self.udp_port = udp_port
    self.serial_port = serial_port
    self.baud = baud
    self.interval = interval
    
    self.relevant_for_listeners = relevant_for_listeners # u, 0
    
  def generate_path(self, mod):
    pth = self.sites_path
    pth = pth + "/" + self.sname + "/ext"
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + mod
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/hysteresis"
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + self.name 
    if not os.path.exists(pth):
      os.mkdir(pth)
    return pth
  
  def histcmd(self, ser, cmnd, resp = '', to = 0.5):
    if ser.timeout != to:
      ser.timeout = to
    ser.flushInput()
    ser.write((cmnd + '\n').encode())
    r = []
    while True:
      l = ser.readline().rstrip().decode()
      if l == '': # EOF
        return r
      r.append(l)
      if len(resp) > 0 and l.find(resp) != -1 or len(r) > 10: # resp found or safeguard
        return r
  
  def dict2confstr(self, section, d):
    r = '[' + section + ']' + '\n'
    for k,v in d.items():
      r += k + '=' + v + '\n'
    return r
  
  def run(self):
    self.alive = True
    self.end = False
    try:
      ser = serial.Serial(self.serial_port, self.baud)
      # r = self.histcmd(ser, 'PWRCHGLOG')
      try:
        while True:
          if self.end:
            ser.close()
            self.alive = False
            log(self.logger, 'Thread closed correctly.', sev=logging.DEBUG)
            return

          d = {'ts': time.strftime('%Y-%m-%d-%H-%M-%S')}

          r = self.histcmd(ser, 'STAT')
          for p in r:
            parts = p.split("=")
            if parts[0] == "Ub__V" and "u" in self.relevant_for_listeners.lower() or parts[0] == "sw_ver":
              try:
                d[parts[0]] = parts[1]
                #maybe create method
                if self.sname != "" and parts[0] == "Ub__V":
                  pth = self.generate_path("voltage")
                  for key, mod in self.downstream_modules_dict.items():
                    mod.set_upstream_info(self.get_name(), "voltage_path", pth)
                  date = datetime.datetime.now().strftime('%Y-%m-%d')
                  fd = open(pth + "/" + date + ".csv", "a+")
                  header = "ts\tUb__V"
                  if not os.path.exists(pth + "/" + date + ".csv") or os.path.getsize(pth + "/" + date + ".csv") == 0:
                    fd.write(header + "\n")
                  tm = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
                  to_write = tm + "\t" + str(parts[1])
                  fd.write(to_write + "\n")
                  fd.close()
                #-------------------
              except:
                log_exception(self.logger, 'Can\'t assign to dict:')
          for key, mod in self.downstream_modules_dict.items():
            try:
              mod.set_upstream_info(self.get_name(), "hist_firmware_version", d['sw_ver'])
              #mod.set_hist_fw(d['sw_ver'])
            except:
              log_exception(self.logger, 'Failed to set sw_ver:')
            try:
              mod.set_upstream_info(self.get_name(), "siwim_system_voltage", d['Ub__V'])
              #mod.set_voltage(d['Ub__V'])
            except:
              log_exception(self.logger, 'Failed to set V:')
            
          msg = self.dict2confstr('mydevice', d)
          sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
          try:
            sock.sendto(msg.encode(), (self.udp_ip, self.udp_port))
          finally:
            sock.close()
          self.zzzzz(self.interval)
      finally:
        ser.close()
    except serial.SerialException as e:
      log(self.logger, e, sev=logging.WARNING)
    except:
      log_exception(self.logger, 'Fatal error:')
      self.alive = False

